/*
 * Copyright (c) 2005 Krister Walfridsson <kristerw@NetBSD.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netinet/ip_mroute.h>
#include <sys/param.h>
#include <sys/sem.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <sys/ttycom.h>
#include <sys/ucred.h>
#include <sys/mount.h>
#include <sys/mtio.h>

/* XXX Need to autoconfigure this... */
#if __NetBSD_Version__ >= 200040000
#include <sys/statvfs.h>
typedef struct { int32_t val[2]; } my_fsid_t;
struct statfs {
  short   f_type;
  u_short f_oflags;
  long    f_bsize;
  long    f_iosize;
  long    f_blocks;
  long    f_bfree;
  long    f_bavail;
  long    f_files;
  long    f_ffree;
  my_fsid_t  f_fsid;
  uid_t   f_owner;
  long    f_flags;
  long    f_syncwrites;
  long    f_asyncwrites;
  long    f_spare[1];
  char    f_fstypename[16];
  char    f_mntonname[90];
  char    f_mntfromname[90];
};
#else
#define HAVE_STATFS 1
#endif

#include <ctype.h>
#include <dirent.h>
#include <elf.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "netbsd_emu.h"

extern char **environ;

static char *filesystem_path;

#define TRACE_FILE_NAME "/tmp/kwa"

#define TRACE_CALL 1
#define TRACE_RET  2
static int trace_flags = 0;

static void trace_call(const char* name, int nofargs);
static void trace_ret32(int32_t x);
static void trace_ret32_2(int32_t x1, int32_t x);
static void trace_ret64(int64_t x);

const char *netbsd_prog;
const char *netbsd_real_prog;
static const char* current_call;

int __getcwd(char *, size_t);
int __getlogin(char *, size_t);
int __posix_rename(const char *, const char *);

static unsigned int membase = 0x20038000;

static struct 
{
  EMUL_ADDR handler;
  int flags;
} signals[32];


/* sigaltstack14 */
EMUL_ADDR ss_sp = 0;
size_t    ss_size = 0;
int       ss_flags = SS_DISABLE;


static EMUL_ADDR pending_sig_handler;
static int pending_sig = -1;
static int running_signal = 0;

static EMUL_ADDR sigcode;

static ARMword stored_cpsr;
static ARMword stored_regs[16];

void
KWARMul_SetPage(ARMul_State * state, ARMword address, void* ptr);

struct
{
  void* addr;
  EMUL_ADDR eaddr;
  EMUL_ADDR neaddr;
  size_t len;
} mmap_table[16];

static char* read_string(EMUL_ADDR str);

static int
xasprintf(char **ret, const char *format, ...) 
  __attribute__((__format__(__printf__, 2, 3)));

static int
xasprintf(char **ret, const char *format, ...)
{
  int status;
  va_list args;

  va_start(args, format);
  status = vasprintf(ret, format, args);
  va_end(args);

  if (*ret == NULL)
    {
      fprintf(stderr, "No memory!\n");
      exit(1);
    }

  return status;
}

static void
trace_call(const char* name, int nofargs)
{
  int i;
  FILE* f;

  current_call = name;

  if ((trace_flags & TRACE_CALL) == 0)
    return;

  f = fopen(TRACE_FILE_NAME, "a+");
  if (f != NULL)
    {
      fprintf(f, "%6d %-8s CALL  %s(", getpid(), netbsd_prog, name);
      for (i = 0; i < nofargs; i++)
        fprintf(f, "0x%x%s", get_parameter(i), i < nofargs - 1 ? "," : "");

      fprintf(f, ")\n");

      fclose(f);
    }
}


static void
trace_ret32(int32_t x)
{
  FILE* f;

  if ((trace_flags & TRACE_RET) == 0)
    return;

  f = fopen(TRACE_FILE_NAME, "a+");
  if (f != NULL)
    {
      fprintf(f, "%6d %-8s RET   %s ", getpid(), netbsd_prog, current_call);
      if (errno == 0)
        fprintf(f, "0x%x\n", x);
      else
        fprintf(f, "-1 %d\n", errno);

      fclose(f);
    }
}


static void
trace_ret32_2(int32_t x1, int32_t x)
{
  FILE* f;

  if ((trace_flags & TRACE_RET) == 0)
    return;

  f = fopen(TRACE_FILE_NAME, "a+");
  if (f != NULL)
    {
      fprintf(f, "%6d %-8s RET   %s ", getpid(), netbsd_prog, current_call);
      if (errno == 0)
        fprintf(f, "0x%x\n", x);
      else
        fprintf(f, "-1 %d\n", errno);

      fclose(f);
    }
}


static void
trace_ret64(int64_t x)
{
  FILE* f;

  if ((trace_flags & TRACE_RET) == 0)
    return;

  f = fopen(TRACE_FILE_NAME, "a+");
  if (f != NULL)
    {
      fprintf(f, "%6d %-8s RET   %s ", getpid(), netbsd_prog, current_call);
      if (errno == 0)
        fprintf(f, "0x%llx\n", (unsigned long long)x);
      else
        fprintf(f, "-1 %d\n", errno);

      fclose(f);
    }
}


void
xlog(const char* format, ...)
{
  FILE* f;
  va_list args;

  f = fopen(TRACE_FILE_NAME, "a+");
  if (f != NULL)
    {
      time_t t = time(NULL);
      va_start(args, format);
      fprintf(f, "%.24s %s: ", ctime(&t), netbsd_prog);
      vfprintf(f, format, args);
      va_end(args);

      fclose(f);
    }
}


static void
xfatal(const char* format, ...)
{
  FILE* f;
  va_list args;

  f = fopen(TRACE_FILE_NAME, "a+");
  if (f != NULL)
    {
      time_t t = time(NULL);
      va_start(args, format);
      fprintf(f, "%.24s %s: ", ctime(&t), netbsd_prog);
      vfprintf(f, format, args);
      va_end(args);

      fclose(f);
    }

  exit(1);
}


static void*
xmalloc(size_t size)
{
  void *p = malloc(size);
  if (p == NULL)
    xfatal("No memory!\n");

  return p;
}


static void*
xrealloc(void *ptr, size_t size)
{
  void *p = realloc(ptr, size);
  if (p == NULL)
    xfatal("No memory!\n");

  return p;
}


static char*
xstrdup(char* s)
{
  void *p = strdup(s);
  if (p == NULL)
    xfatal("No memory!\n");

  return p;
}


static char*
read_path(EMUL_ADDR path)
{
  return target_to_host_path(read_string(path));
}


static void
write_path(EMUL_ADDR dst, char* path)
{
  const char *p = host_to_target_path(path);

  do
    {
      WRITE_INT8(dst++, *p);
    } while (*p++ != '\0');
}


/* Read a string from the target's address space. */
static char*
read_string(EMUL_ADDR str)
{
  static void *buf = NULL;

  EMUL_ADDR  tmp_str = str;
  while (READ_INT8(tmp_str++) !=  '\0')
    ;

  buf = xrealloc(buf, tmp_str - str);
  read_memory(buf, str, tmp_str - str);

  return buf;
}


static void
read_struct_flock(struct flock *f, EMUL_ADDR addr)
{
  f->l_start = read_int64(addr); addr += 8;
  f->l_len = read_int64(addr); addr += 8;
  f->l_pid = READ_INT32(addr); addr += 4;
  f->l_type = read_int16(addr); addr += 2;
  f->l_whence = read_int16(addr);
}


static EMUL_ADDR
read_struct_timeval(struct timeval *ts, EMUL_ADDR addr)
{
  ts->tv_sec = read_long(addr);  addr += EMUL_SIZEOF_LONG;
  ts->tv_usec = read_long(addr); addr += EMUL_SIZEOF_LONG;

  return addr;
}


static EMUL_ADDR
write_struct_timeval(EMUL_ADDR addr, struct timeval *ts)
{
  write_long(addr, ts->tv_sec);  addr += EMUL_SIZEOF_LONG;
  write_long(addr, ts->tv_usec); addr += EMUL_SIZEOF_LONG;

  return addr;
}


static void
write_struct_timezone(EMUL_ADDR addr, struct timezone *tz)
{
  WRITE_INT32(addr    , tz->tz_minuteswest);
  WRITE_INT32(addr + 4, tz->tz_dsttime);
}


static void
read_struct_timespec(struct timespec *dst, EMUL_ADDR src)
{
  dst->tv_sec = read_long(src); src += EMUL_SIZEOF_LONG;
  dst->tv_nsec = read_long(src); src += EMUL_SIZEOF_LONG;
}


static void
write_struct_timespec(EMUL_ADDR dst, struct timespec *src)
{
  write_long(dst, src->tv_sec);  dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->tv_nsec);  dst += EMUL_SIZEOF_LONG;
}


static void
write_struct_stat(EMUL_ADDR dst, struct stat *src)
{
  WRITE_INT32(dst, src->st_dev); dst += 4;
  WRITE_INT32(dst, src->st_ino); dst += 4;
  WRITE_INT32(dst, src->st_mode); dst += 4;
  WRITE_INT32(dst, src->st_nlink); dst += 4;
  WRITE_INT32(dst, src->st_uid); dst += 4;
  WRITE_INT32(dst, src->st_gid); dst += 4;
  WRITE_INT32(dst, src->st_rdev); dst += 4;
  WRITE_INT32(dst, src->st_atimespec.tv_sec); dst += 4;
  WRITE_LONG(dst, src->st_atimespec.tv_nsec); dst += EMUL_SIZEOF_LONG;
  WRITE_INT32(dst, src->st_mtimespec.tv_sec); dst += 4;
  WRITE_LONG(dst, src->st_mtimespec.tv_nsec); dst += EMUL_SIZEOF_LONG;
  WRITE_INT32(dst, src->st_ctimespec.tv_sec); dst += 4;
  WRITE_LONG(dst, src->st_ctimespec.tv_nsec); dst += EMUL_SIZEOF_LONG;
  write_int64(dst, src->st_size); dst += 8;
  write_int64(dst, src->st_blocks); dst += 8;
  WRITE_INT32(dst, src->st_blksize);  dst += 4;
  WRITE_INT32(dst, src->st_flags);  dst += 4;
  WRITE_INT32(dst, src->st_gen); dst += 4;
#if 0  /* XXX */
  write_int64(dst, src->st_qspare[0]); dst += 8;
  write_int64(dst, src->st_qspare[1]);
#endif
}


static void
write_struct_statfs(EMUL_ADDR dst, struct statfs *src)
{
  WRITE_INT16(dst, src->f_type); dst += 2;
  WRITE_INT16(dst, src->f_oflags); dst += 2;
  WRITE_LONG(dst, src->f_bsize); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_iosize); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_blocks); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_bfree); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_bavail); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_files); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_ffree); dst += EMUL_SIZEOF_LONG;
  WRITE_INT32(dst, src->f_fsid.val[0]); dst += 4;
  WRITE_INT32(dst, src->f_fsid.val[1]); dst += 4;
  WRITE_INT32(dst, src->f_owner); dst += 4;  
  WRITE_LONG(dst, src->f_flags); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_syncwrites); dst += EMUL_SIZEOF_LONG;
  WRITE_LONG(dst, src->f_asyncwrites); dst += EMUL_SIZEOF_LONG;
  /* f_spare[1]; */ dst += EMUL_SIZEOF_LONG;
  write_memory(dst, src->f_fstypename, MFSNAMELEN + MNAMELEN + MNAMELEN);
}


static void
write_struct_rusage(EMUL_ADDR dst, struct rusage *src)
{
  dst = write_struct_timeval(dst, &src->ru_utime);
  dst = write_struct_timeval(dst, &src->ru_stime);
  write_long(dst, src->ru_maxrss); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_ixrss); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_isrss); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_isrss); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_minflt); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_majflt); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_nswap); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_inblock); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_oublock); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_msgsnd); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_msgrcv); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_nsignals); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_nvcsw); dst += EMUL_SIZEOF_LONG;
  write_long(dst, src->ru_nivcsw); dst += EMUL_SIZEOF_LONG;
}


static void
do_exit(void) /* XXX */
{
  int status = get_parameter(0);

  exit(status);
}


static void
do_fork(void)
{
  int status = fork();
  if (status == 0)
    write_status32_2(getppid(), 1);  /* Child */
  else
    write_status32_2(status, 0);  /* Parent */
}


static void
do_read(void)
{
  int d = get_parameter(0);
  EMUL_ADDR buf = get_parameter(1);
  size_t nbytes = get_parameter(2);

  void* tmp_buf = xmalloc(nbytes);

  int status = read(d, tmp_buf, nbytes);

  write_status32(status);
  if (status != -1)
    write_memory(buf, tmp_buf, status);

  free(tmp_buf);
}


static void
do_write(void)
{
  int status;

  int d = get_parameter(0);
  EMUL_ADDR buf = get_parameter(1);
  size_t nbytes = get_parameter(2);

  /* Make sure we get a valid buffer even if nbytes == 0. */
  void* tmp_buf = xmalloc(MAX(nbytes, 1));
  read_memory(tmp_buf, buf, nbytes);

  status = write(d, tmp_buf, nbytes);
  write_status32(status);

  free(tmp_buf);
}


static void
do_open(void)
{
  EMUL_ADDR path = get_parameter(0);
  int flags = get_parameter(1);
  mode_t mode = get_parameter(2);

  int status = open(read_path(path), flags, mode);

  write_status32(status);
}


static void
do_close(void)
{
  int d = get_parameter(0);

  int status = close(d);

  write_status32(status);
}


static void
do_wait4(void)  /* XXX */
{
  pid_t wpid = get_parameter(0);
  EMUL_ADDR statusp =  get_parameter(1);
  int options =  get_parameter(2);
  EMUL_ADDR rusage =  get_parameter(3);

  int status;

  int tmp_status;
  struct rusage buf;
  void *buf_ptr = rusage != 0 ? &buf : NULL;

#if 0
  printf("WAITING FOR %d\n", wpid);
#endif

  status = wait4(wpid, &tmp_status, options, buf_ptr);

  write_status32(status);
  if (status != -1)
    {
      WRITE_INT32(statusp, tmp_status);

      if (rusage != 0)
        write_struct_rusage(rusage, &buf);
    }
}


static void
do_link(void)
{
  EMUL_ADDR name1 = get_parameter(0);
  EMUL_ADDR name2 = get_parameter(1);

  /* read_path() returns a static bufer, so we cannot call it for both
   * parameters in the function call. */
  char* tmp_name1 = xstrdup(read_path(name1));
  int status = link(tmp_name1, read_path(name2));
  free(tmp_name1);

  write_status32(status);
}


static void
do_unlink(void)
{
  EMUL_ADDR path = get_parameter(0);

  int status = unlink(read_path(path));

  write_status32(status);
}


static void
do_chdir(void)
{
  EMUL_ADDR path = get_parameter(0);

  int status = chdir(read_path(path));

  write_status32(status);
}


static void
do_fchdir(void)
{
  int fd = get_parameter(0);

  int status = fchdir(fd);

  write_status32(status);
}


static void
do_mknod(void)
{
  EMUL_ADDR path = get_parameter(0);
  mode_t mode = get_parameter(1);
  dev_t dev = get_parameter(2);

  int status = mknod(read_path(path), mode, dev);

  write_status32(status);
}


static void
do_chmod(void)
{
  EMUL_ADDR path = get_parameter(0);
  mode_t mode = get_parameter(1);

  int status = chmod(read_path(path), mode);

  write_status32(status);
}


static void
do_chown(void)
{
  EMUL_ADDR path = get_parameter(0);
  uid_t owner = get_parameter(1);
  gid_t group = get_parameter(2);

  int status = chown(read_path(path), owner, group);

  write_status32(status);
}


static void
do_break(void)  /* XXX */
{
#if 0
  EMUL_ADDR new_break = get_parameter(0);

  fprintf(tfile, "New break = 0x%x\n", new_break);
#endif
  /* XXX Return succes for now -- we do not have any SEGV handling
   * anyway. */
  write_status32(0);
}


static void
do_getfsstat(void)  /* XXX */
{
#ifdef HAVE_STATFS
  int status;

  EMUL_ADDR buf = get_parameter(0);
  long bufsize = get_parameter(1);
  int flags = get_parameter(2);

  void* tmp_buf = NULL;
  if (buf != 0)
    tmp_buf = xmalloc(bufsize);

  status = getfsstat(tmp_buf, bufsize, flags);
  write_status32(status);
  if (status != -1 && buf != 0)
    write_memory(buf, tmp_buf, bufsize);

  free(tmp_buf);
#endif
}


static void
do_getpid(void)
{
  int status = getpid();

  write_status32(status);
}


static void
do_setuid(void)
{
  uid_t uid = get_parameter(0);

  int status = setuid(uid);
  write_status32(status);
}


static void
do_getuid(void)
{
  int status = getuid();

  write_status32(status);
}


static void
do_geteuid(void)
{
  int status = geteuid();

  write_status32(status);
}


static void
do_recvfrom(void)
{
  ssize_t status;

  int s = get_parameter(0);
  EMUL_ADDR buf = get_parameter(1);
  size_t len = get_parameter(2);
  int flags = get_parameter(3);
  EMUL_ADDR from = get_parameter(4);
  EMUL_ADDR fromlen = get_parameter(5);

  void *tmp_buf = NULL;
  void *tmp_from = NULL;
  void *tmp_fromlen = NULL;
  socklen_t tmp_flen = 0;

  if (buf != 0)
    tmp_buf = xmalloc(len);
  if (fromlen != 0)
    {
      tmp_flen = READ_INT32(fromlen);
      tmp_fromlen = &tmp_flen;
    }
  if (from != 0)
    tmp_from = xmalloc(tmp_flen + 1);  /* +1 to handle == 0. */

  status = recvfrom(s, tmp_buf, len, flags, tmp_from, tmp_fromlen);
  if (status != -1)
    {
      write_memory(buf, tmp_buf, len);
      if (tmp_from != NULL)
        write_memory(from, tmp_from, tmp_flen);
    }

  free(tmp_buf);
  free(tmp_from);

  write_status32(status);
}


static void
do_access(void)
{
  EMUL_ADDR path = get_parameter(0);
  int mode = get_parameter(1);

  int status = access(read_path(path), mode);

  write_status32(status);
}


static void
do_chflags(void)
{
  EMUL_ADDR path = get_parameter(0);
  u_long flags = get_parameter(1);

  int status = chflags(read_path(path), flags);

  write_status32(status);
}


static void
do_fchflags(void)
{
  int fd = get_parameter(0);
  u_long flags = get_parameter(1);

  int status = fchflags(fd, flags);

  write_status32(status);
}


static void
do_sync(void)
{
  sync();
  write_status32(0);  /* XXX ? */
}


static void
do_kill(void)
{
  pid_t pid = get_parameter(0);
  int sig = get_parameter(1);

  int status = kill(pid, sig);

  write_status32(status);
}


static void
do_getppid(void)
{
  int status = getppid();

  write_status32(status);
}


static void
do_dup(void)
{
  int oldd = get_parameter(0);

  int status = dup(oldd);

  write_status32(status);
}


static void
do_pipe(void)
{
  int status;

  int fildes[2];

  status = pipe(fildes);

  write_status32_2(fildes[0], fildes[1]);
}


static void
do_getegid(void)
{
  int status = getegid();

  write_status32(status);
}


static void
do_getgid(void)
{
  int status = getgid();

  write_status32(status);
}


static void
do___getlogin(void)  /* XXX */
{
  EMUL_ADDR name= get_parameter(0);
  size_t buflen= get_parameter(1);

  void* tmp_name = xmalloc(buflen);
  int status = __getlogin(tmp_name, buflen);
  if (status != -1)
    write_memory(name, tmp_name, buflen);

  write_status32(status);

  free(tmp_name);
}


static void
do_ioctl(void)  /* XXX */
{
  int status;

  int d = get_parameter(0);
  unsigned long request = get_parameter(1);
  EMUL_ADDR argp = get_parameter(2);

  if (request == TIOCGETA)
    {
      struct termios buf;
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        write_memory(argp, &buf, sizeof(buf)); /* XXX */
    }
  else if (request == TIOCSETA)
    {
      struct termios buf;
      read_memory(&buf, argp, sizeof(buf));
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == TIOCSETAW)
    {
      struct termios buf;
      read_memory(&buf, argp, sizeof(buf));
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == TIOCSETAF)
    {
      struct termios buf;
      read_memory(&buf, argp, sizeof(buf));
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == FIOCLEX)
    {
      status = ioctl(d, request, NULL);
      write_status32(status);
    }
  else if (request == FIONCLEX)
    {
      status = ioctl(d, request, NULL);
      write_status32(status);
    }
  else if (request == FIONREAD)
    {
      int buf;
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        WRITE_INT32(argp, buf);
    }
  else if (request == FIONBIO)
    {
      int buf = READ_INT32(argp);  /* XXX argp may be NULL? */
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == FIOASYNC)
    {
      int buf = READ_INT32(argp);  /* XXX argp may be NULL? */
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == FIOSETOWN)
    {
      int buf = READ_INT32(argp);
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == FIOGETOWN)
    {
      int buf;
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        WRITE_INT32(argp, buf);
    }
  else if (request == TIOCGPGRP)
    {
      int buf;
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        WRITE_INT32(argp, buf);
    }
  else if (request == TIOCSPGRP)
    {
      int s = READ_INT32(argp);
      status = ioctl(d, request, &s);
      write_status32(status);
    }
  else if (request == TIOCEXT)
    {
      int s = READ_INT32(argp);
      status = ioctl(d, request, &s);
      write_status32(status);
    }
  else if (request == TIOCSTART)
    {
      status = ioctl(d, request);
      write_status32(status);
    }
  else if (request == TIOCSTOP)
    {
      status = ioctl(d, request);
      write_status32(status);
    }
  else if (request == TIOCSETD)
    {
      int buf = READ_INT32(argp);
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == TIOCGETD)
    {
      int buf;
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        WRITE_INT32(argp, buf);
    }
  else if (request == TIOCGWINSZ)
    {
      struct winsize buf;
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        write_memory(argp, &buf, sizeof(buf)); /* XXX */
    }
  else if (request == TIOCSETAW)
    {
      struct termios buf;
      read_memory(&buf, argp, sizeof(buf));
      status = ioctl(d, request, &buf);
      write_status32(status);
    }
  else if (request == MTIOCGET)
    {
      struct mtget buf;
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        write_memory(argp, &buf, sizeof(buf)); /* XXX */
    }
  else if (request == SIOCGETSGCNT)
    {
      struct sioc_sg_req buf;
      read_memory(&buf, argp, sizeof(buf));
      status = ioctl(d, request, &buf);
      write_status32(status);
      if (status != -1)
        write_memory(argp, &buf, sizeof(buf)); /* XXX */
    }
  else
    xfatal("Unimplemented IOCTL 0x%lx\n", request);
}


static void
do_symlink(void)
{
  EMUL_ADDR name1 = get_parameter(0);
  EMUL_ADDR name2 = get_parameter(1);

  /* read_path() returns a static bufer, so we cannot call it for both
   * parameters in the function call. */
  char* tmp_name1 = xstrdup(read_path(name1));
  int status = symlink(tmp_name1, read_path(name2));
  free(tmp_name1);

  write_status32(status);
}


static void
do_readlink(void)
{
  EMUL_ADDR path = get_parameter(0);
  EMUL_ADDR buf = get_parameter(1);
  size_t bufsiz = get_parameter(2);

  /* Need bigger buffer, since path may start with "filesystem_path". */
  size_t tmp_bufsize = bufsiz + strlen(filesystem_path) + 1;
  char* tmp_buf = xmalloc(tmp_bufsize);

  int status = readlink(read_path(path), tmp_buf, tmp_bufsize);
  if (status != -1)
    {
      const char* p;

      tmp_buf[status] = '\0';
      p = host_to_target_path(tmp_buf);
      status = MIN(strlen(p), bufsiz);
      write_memory(buf, p, status);
    }
  write_status32(status);

  free(tmp_buf);
}


static void
do_execve(void)  /* XXX */
{
  int status;
  int i;
  EMUL_ADDR p;

  EMUL_ADDR path = get_parameter(0);
  EMUL_ADDR argv = get_parameter(1);
  EMUL_ADDR envp = get_parameter(2);

  int argc, envc;

  char **tmp_argv;
  char **tmp_env;

  for (argc = 0, p = argv; READ_INT32(p) != 0; argc++, p += 4)
    ;
  tmp_argv = xmalloc((argc + 1) * sizeof(char*));

  for (envc = 0, p = envp; READ_INT32(p) != 0; envc++, p += 4)
    ;
  tmp_env = xmalloc((envc + 1) * sizeof(char*));

  for (i = 0, p = argv; i < argc; i++, p += 4)
    tmp_argv[i] = xstrdup(read_string(READ_INT32(p)));
  tmp_argv[i] = NULL;

  /* The environment the emulated progam will see. */
  for (i = 0, p = envp; i < envc; i++, p += 4)
    tmp_env[i] = xstrdup(read_string(READ_INT32(p)));
  tmp_env[i] = NULL;

  /* This does not return if successful. */
  status = run_execve(xstrdup(read_string(path)), tmp_argv, tmp_env);

  write_status32(status);

  for (i = 0; tmp_argv[i] != NULL; i++)
    free(tmp_argv[i]);
  free(tmp_argv);
  for (i = 0; tmp_env[i] != NULL; i++)
    free(tmp_env[i]);
  free(tmp_env);
}


static void
do_umask(void)
{
  mode_t numask = get_parameter(0);

  mode_t status = umask(numask);
  write_status32(status);  
}


static void
do_vfork(void)
{
  /* Emulating vfork is a bit tricky, but most applications does not
   * seem to care if fork is called instead... */
  do_fork();
}


static void
do_vadvise(void)  /* XXX */
{
  write_status32(0);
}


static void
do_munmap(void)  /* XXX */
{
  int status;

  EMUL_ADDR addr = get_parameter(0);
  size_t len = get_parameter(1);
#if 0
  if (mmap_table[0].eaddr != addr || mmap_table[0].len != len)
    {
      fprintf(tfile, "XXX munmap 0x%x 0x%x\n", addr, len);
    }
#endif
#if 0
  status = munmap(mmap_table[0].addr, len);
  if (status != -1)
    {
      mmap_table[0].addr = NULL;
      mmap_table[0].eaddr = 0;
      mmap_table[0].len = 0;

      KWARMul_SetPage(emul_state, addr, 0);
    }
#else
  if (mmap_table[0].eaddr == addr
      && mmap_table[0].len == len
      && mmap_table[0].neaddr == membase)
    {
      mmap_table[0].eaddr = 0;
      mmap_table[0].len = 0;
      mmap_table[0].neaddr = 0;

      membase = addr;
    }

  status = 0;
#endif
  write_status32(status);
}


static void
do_mprotect(void)
{
  /* XXX There is no sensible thing to do in this syscall. */
  write_status32(0);
}


static void
do_madvise(void)
{
  /* There is no sensible thing to do in this syscall.  Just ignore it. */
  write_status32(0);
}


static void
do_getgroups(void)
{
  int status;

  int gidsetlen = get_parameter(0);
  EMUL_ADDR gidset = get_parameter(1);

  gid_t *tmp_gidset = NULL;

  if (gidsetlen > 0)
    tmp_gidset = xmalloc(gidsetlen * sizeof(gid_t));

  status = getgroups(gidsetlen, tmp_gidset);
  write_status32(status);
  if (status != -1 && tmp_gidset != 0)
    write_memory(gidset, tmp_gidset, status * sizeof(gid_t)); /* XXX */

  free(tmp_gidset);
}


static void
do_setgroups(void)
{
  int status;

  int gidsetlen = get_parameter(0);
  EMUL_ADDR gidset = get_parameter(1);

  gid_t *tmp_gidset = NULL;

  if (gidsetlen > 0)  /* XXX What if lem == 0? */
    {
      tmp_gidset = xmalloc(gidsetlen * sizeof(gid_t));
      read_memory(tmp_gidset, gidset, gidsetlen * sizeof(gid_t)); /* XXX */
    }

  status = setgroups(gidsetlen, tmp_gidset);
  write_status32(status);

  free(tmp_gidset);
}


static void
do_getpgrp(void)
{
  int status = getpgrp();

  write_status32(status);
}


static void
do_setpgid(void)
{
  pid_t pid = get_parameter(0);
  pid_t pgrp = get_parameter(1);

  int status = setpgid(pid, pgrp);

  write_status32(status);
}


static void
do_setitimer(void)
{
  int status;

  int which = get_parameter(0);
  EMUL_ADDR value = get_parameter(1);
  EMUL_ADDR ovalue = get_parameter(2);

  struct itimerval buf_value;
  struct itimerval buf_ovalue;

  struct itimerval *tmp_ovalue = NULL;

  EMUL_ADDR a = read_struct_timeval(&buf_value.it_interval, value);
  (void)read_struct_timeval(&buf_value.it_value, a);

  if (ovalue != 0)
    tmp_ovalue = &buf_ovalue;

  status = setitimer(which, &buf_value, tmp_ovalue);
  if (status != -1 && tmp_ovalue != NULL)
    {
      a = write_struct_timeval(ovalue, &buf_ovalue.it_interval);
      (void)write_struct_timeval(a, &buf_ovalue.it_value);
    }

  write_status32(status);
}


static void
do_dup2(void)
{
  int oldd = get_parameter(0);
  int newd = get_parameter(1);

  int status = dup2(oldd, newd);

  write_status32(status);
}


static void
do_fcntl(void)
{
  int status;

  int fd = get_parameter(0);
  int cmd = get_parameter(1);
  EMUL_ADDR arg = get_parameter(2);

  switch (cmd)
    {
    case F_DUPFD:
    case F_GETFD:
    case F_SETFD:
    case F_GETFL:
    case F_SETFL:
    case F_GETOWN:
    case F_SETOWN:
      status = fcntl(fd, cmd, arg);
      break;

    case F_GETLK:
    case F_SETLK:
    case F_SETLKW:
      {
        struct flock tmp_arg;
        
        read_struct_flock(&tmp_arg, arg);
        status = fcntl(fd, cmd, tmp_arg);
      }
      break;

    default:
      errno = EINVAL;
      status = -1;
    }

  write_status32(status);
}


static void
do_select(void)  /* XXX */
{
  int status;

  int nfds = get_parameter(0);
  EMUL_ADDR readfds = get_parameter(1);
  EMUL_ADDR writefds = get_parameter(2);
  EMUL_ADDR exceptfds = get_parameter(3);
  EMUL_ADDR timeout = get_parameter(4);

  fd_set tmp_readfds, *tmp_readfds_ptr = NULL;
  fd_set tmp_writefds, *tmp_writefds_ptr = NULL;
  fd_set tmp_exceptfds, *tmp_exceptfds_ptr = NULL;
  struct timeval tmp_timeout, *tmp_timeout_ptr = NULL;

  /* XXX These sizes needs to be checked. */
  if (readfds != 0)
    {
      read_memory(&tmp_readfds, readfds, sizeof(tmp_readfds));
      tmp_readfds_ptr = &tmp_readfds;
    }
  if (writefds != 0)
    {
      read_memory(&tmp_writefds, writefds, sizeof(tmp_writefds));
      tmp_writefds_ptr = &tmp_writefds;
    }
  if (exceptfds != 0)
    {
      read_memory(&tmp_exceptfds, exceptfds, sizeof(tmp_exceptfds));
      tmp_exceptfds_ptr = &tmp_exceptfds;
    }

  if (timeout != 0)
    {
      read_memory(&tmp_timeout, timeout, sizeof(tmp_timeout));
      tmp_timeout_ptr = &tmp_timeout;
    }

  status = select(nfds, tmp_readfds_ptr, tmp_writefds_ptr,
                  tmp_exceptfds_ptr, tmp_timeout_ptr);

  write_status32(status);
  if (status != -1)
    {
      /* XXX These sizes needs to be checked. */
      if (readfds != 0)
        write_memory(readfds, &tmp_readfds, sizeof(tmp_readfds));
      if (writefds != 0)
        write_memory(writefds, &tmp_writefds, sizeof(tmp_writefds));
      if (exceptfds != 0)
        write_memory(exceptfds, &tmp_exceptfds, sizeof(tmp_exceptfds));
    }
}


static void
do_fsync(void)
{
  int fd = get_parameter(0);

  int status = fsync(fd);

  write_status32(status);
}


static void
do_setpriority(void)
{
  int which = get_parameter(0);
  int who = get_parameter(1);
  int prio = get_parameter(2);

  int status = setpriority(which, who, prio);

  write_status32(status);
}


static void
do_socket(void)
{
  int domain = get_parameter(0);
  int type = get_parameter(1);
  int protocol = get_parameter(2);

  int status = socket(domain, type, protocol);

  write_status32(status);
}


static void
do_connect(void)  /* XXX */
{
  int status;

  int s = get_parameter(0);
  EMUL_ADDR name = get_parameter(1);
  socklen_t namelen = get_parameter(2);

  void* tmp_buf = xmalloc(namelen);
  read_memory(tmp_buf, name, namelen); /* XXX endianness */

  status = connect(s, tmp_buf, namelen);

  write_status32(status);

  free(tmp_buf);
}


static void
do_getpriority(void)
{
  int which = get_parameter(0);
  int who = get_parameter(1);

  int status = getpriority(which, who);

  write_status32(status);
}


static void
do_bind(void)  /* XXX */
{
  int status;

  int s = get_parameter(0);
  EMUL_ADDR name = get_parameter(1);
  socklen_t namelen = get_parameter(2);

  void* tmp_name = xmalloc(namelen);
  read_memory(tmp_name, name, namelen);

  status = bind(s, tmp_name, namelen);

  write_status32(status);
}


static void
do_setsockopt(void)  /* XXX */
{
  int s = get_parameter(0);
  int level = get_parameter(1);
  int optname = get_parameter(2);
  EMUL_ADDR optval = get_parameter(3);
  socklen_t optlen = get_parameter(4);

  xfatal("syscall setsockopt(0x%x, 0x%x, 0x%x, 0x%x, 0x%x) not supported\n",
         s, level, optname, optval, optlen);
}


static void
do_listen(void)
{
  int s = get_parameter(0);
  int backlog = get_parameter(1);

  int status = listen(s, backlog);

  write_status32(status);
}


static void
do_gettimeofday(void)
{
  EMUL_ADDR tp = get_parameter(0);
  EMUL_ADDR tzp = get_parameter(1);
  struct timeval t;
  struct timezone tz;

  int status = gettimeofday(tp != 0 ? &t : NULL, tzp != 0 ? &tz : NULL);
  write_status32(status);
  if (status != -1)
    {
      if (tp != 0)
        (void)write_struct_timeval(tp, &t);
      if (tzp != 0)
        write_struct_timezone(tzp, &tz);
    }
}


static void
do_getrusage(void)
{
  int who = get_parameter(0);
  EMUL_ADDR rusage = get_parameter(1);
  struct rusage buf;

  int status = getrusage(who, &buf);

  write_status32(status);
  if (status != -1)
    write_struct_rusage(rusage, &buf);
}


static void
do_getsockopt(void)  /* XXX */
{
  int status;

  int s = get_parameter(0);
  int level = get_parameter(1);
  int optname = get_parameter(2);
  EMUL_ADDR optval = get_parameter(3);
  EMUL_ADDR optlen = get_parameter(4);

  socklen_t tmp_len;

  socklen_t *tmp_optval = NULL;

  tmp_len = READ_INT32(optlen);
  if (optval != 0)
    tmp_optval = xmalloc(tmp_len);

  status = getsockopt(s, level, optname, tmp_optval, &tmp_len);

  write_status32(status);
  if (status != -1)
    {
      switch(optname)
        {
        case SO_DEBUG:
        case SO_REUSEADDR:
        case SO_REUSEPORT:
        case SO_KEEPALIVE:
        case SO_DONTROUTE:
        case SO_BROADCAST:
        case SO_OOBINLINE:
        case SO_SNDBUF:
        case SO_RCVBUF:
        case SO_SNDLOWAT:
        case SO_RCVLOWAT:
        case SO_TIMESTAMP:
        case SO_TYPE:
        case SO_ERROR:
          {
            int tmp;
            if (tmp_len != 4)
              xfatal("getsockopt, tmp_len == %d", tmp_len);
            if(tmp_optval == NULL)
              xfatal("getsockopt, tmp_optval == NULL");
            memcpy(&tmp, tmp_optval, 4);
            WRITE_INT32(optval, tmp);
            WRITE_INT32(optlen, tmp_len);
          }
          break;
        case SO_SNDTIMEO:
        case SO_RCVTIMEO:
        case SO_LINGER:
        default:
          xfatal("syscall getsockopt(0x%x, 0x%x, 0x%x, 0x%x, 0x%x)"
                 " not supported\n",
                 s, level, optname, optval, optlen);
        }
    }
}


static void
do_readv(void)  /* XXX */
{
  ssize_t status;
  int i;

  int d = get_parameter(0);
  EMUL_ADDR iov = get_parameter(1);
  int iovcnt = get_parameter(2);

  struct iovec *tmp_iov = xmalloc(sizeof(struct iovec) * iovcnt);

  for (i = 0; i < iovcnt; i++)
    {
      tmp_iov[i].iov_len = READ_INT32(iov + i * 8 + 4);
      tmp_iov[i].iov_base = xmalloc(tmp_iov[i].iov_len);
    }

  status = readv(d, tmp_iov, iovcnt);

  write_status32(status);
  if (status != -1)
    {
      for (i = 0; i < iovcnt; i++)
        {
          write_memory(READ_INT32(iov + i * 8),
                       tmp_iov[i].iov_base, tmp_iov[i].iov_len);
          free(tmp_iov[i].iov_base);
        }
    }

    free(tmp_iov);
}


static void
do_writev(void)  /* XXX */
{
  int status;
  int i;

  int d = get_parameter(0);
  EMUL_ADDR iov = get_parameter(1);
  int iovcnt = get_parameter(2);

  struct iovec *tmp_iov = xmalloc(sizeof(struct iovec) * iovcnt);

  for (i = 0; i < iovcnt; i++)
    {
      tmp_iov[i].iov_len = READ_INT32(iov + 4);
      tmp_iov[i].iov_base = xmalloc(tmp_iov[i].iov_len);
      read_memory(tmp_iov[i].iov_base, READ_INT32(iov), tmp_iov[i].iov_len);
      iov += 8;
    }

  status = writev(d, tmp_iov, iovcnt);
  write_status32(status);

  for (i = 0; i < iovcnt; i++)
    free(tmp_iov[i].iov_base);
  free(tmp_iov);
}


static void
do_fchown(void)
{
  int fd = get_parameter(0);
  uid_t owner = get_parameter(1);
  gid_t group = get_parameter(2);

  int status = fchown(fd, owner, group);

  write_status32(status);
}


static void
do_fchmod(void)
{
  int fd = get_parameter(0);
  mode_t mode = get_parameter(1);

  int status = fchmod(fd, mode);
  write_status32(status);  
}


static void
do_setreuid(void)
{
  uid_t ruid = get_parameter(0);
  uid_t euid = get_parameter(1);

  int status = setreuid(ruid, euid);

  write_status32(status);
}


static void
do_setregid(void)
{
  gid_t rgid = get_parameter(0);
  gid_t egid = get_parameter(1);

  int status = setregid(rgid, egid);

  write_status32(status);
}


static void
do_rename(void)
{
  EMUL_ADDR from = get_parameter(0);
  EMUL_ADDR to = get_parameter(1);

  /* read_path() returns a static bufer, so we cannot call it for both
   * parameters in the function call. */
  char* tmp_from = xstrdup(read_path(from));
  int status = rename(tmp_from, read_path(to));
  free(tmp_from);

  write_status32(status);
}


static void
do_flock(void)
{
  int fd = get_parameter(0);
  int operation = get_parameter(1);

  int status = flock(fd, operation);

  write_status32(status);
}


static void
do_mkfifo(void)
{
  EMUL_ADDR path = get_parameter(0);
  mode_t mode = get_parameter(1);

  int status = mkfifo(read_path(path), mode);

  write_status32(status);
}


static void
do_sendto(void)  /* XXX */
{
  ssize_t status;

  int s = get_parameter(0);
  EMUL_ADDR msg = get_parameter(1);
  size_t len = get_parameter(2);
  int flags = get_parameter(3);
  EMUL_ADDR to = get_parameter(4);
  socklen_t tolen = get_parameter(5);

  void* tmp_msg = xmalloc(len);
  void* tmp_to = xmalloc(tolen);

  read_memory(tmp_msg, msg, len);
  read_memory(tmp_to, to, tolen);

  status = sendto(s, tmp_msg, len, flags, tmp_to, tolen);

  write_status32(status);

  free(tmp_msg);
  free(tmp_to);
}


static void
do_shutdown(void)
{
  int s = get_parameter(0);
  int how = get_parameter(1);

  int status = shutdown(s, how);

  write_status32(status);
}


static void
do_socketpair(void)  /* XXX */
{
  int status;

  int d = get_parameter(0);
  int type = get_parameter(1);
  int protocol = get_parameter(2);
  EMUL_ADDR sv = get_parameter(3);

  int tmp_sv[2];

  status = socketpair(d, type, protocol, tmp_sv);

  write_status32(status);
  if (status != -1)
    {
      WRITE_INT32(sv    , tmp_sv[0]);
      WRITE_INT32(sv + 4, tmp_sv[1]);
    }
}


static void
do_mkdir(void)
{
  EMUL_ADDR path = get_parameter(0);
  mode_t mode = get_parameter(1);

  int status = mkdir(read_path(path), mode);
  write_status32(status);  
}


static void
do_rmdir(void)
{
  EMUL_ADDR path = get_parameter(0);

  int status = rmdir(read_path(path));
  write_status32(status);  
}


static void
do_utimes(void)
{
  int status;

  EMUL_ADDR path = get_parameter(0);
  EMUL_ADDR times = get_parameter(1);

  struct timeval tmp_buf[2];
  void *tmp_times = NULL;
  if (times != 0)
    {
      times = read_struct_timeval(&tmp_buf[0], times);
      (void)  read_struct_timeval(&tmp_buf[1], times);
      tmp_times = &tmp_buf[0];
    }

  status = utimes(read_path(path), tmp_times);
  write_status32(status);  
}


static void
do_setsid(void)
{
  int status = setsid();
  write_status32(status);
}


static void
do_statfs(void)
{
#ifdef HAVE_STATFS
  EMUL_ADDR path = get_parameter(0);
  EMUL_ADDR buf = get_parameter(1);
  struct statfs tmp_buf;

  int status = statfs(read_path(path), &tmp_buf);

  write_status32(status);
  if (status != -1)
    write_struct_statfs(buf, &tmp_buf);
#endif
}


static void
do_fstatfs(void)
{
#ifdef HAVE_STATFS
  int fd = get_parameter(0);
  EMUL_ADDR buf = get_parameter(1);
  struct statfs tmp_buf;

  int status = fstatfs(fd, &tmp_buf);

  write_status32(status);
  if (status != -1)
    write_struct_statfs(buf, &tmp_buf);
#endif
}


static void
do_pread(void)  /* XXX */
{
  int d = get_parameter(0);
  EMUL_ADDR buf  = get_parameter(1);
  size_t nbytes = get_parameter(2);
  off_t offset = ((off_t)get_parameter(5) << 32) | get_parameter(4);

  void* tmp_buf = xmalloc(nbytes);

  ssize_t status = pread(d, tmp_buf, nbytes, offset);

  write_status32(status);
  if (status != -1)
    write_memory(buf, tmp_buf, nbytes);

  free(tmp_buf);
}


static void
do_pwrite(void)  /* XXX */
{
  ssize_t status;

  int d = get_parameter(0);
  EMUL_ADDR buf  = get_parameter(1);
  size_t nbytes = get_parameter(2);
  off_t offset = ((off_t)get_parameter(5) << 32) | get_parameter(4);

  void* tmp_buf = xmalloc(nbytes);
  read_memory(tmp_buf, buf, nbytes);

  status = pwrite(d, tmp_buf, nbytes, offset);
  write_status32(status);

  free(tmp_buf);
}


static void
do_setgid(void)
{
  gid_t gid = get_parameter(0);

  int status = setgid(gid);
  write_status32(status);
}


static void
do_setegid(void)
{
  gid_t gid = get_parameter(0);

  int status = setegid(gid);
  write_status32(status);
}


static void
do_seteuid(void)
{
  uid_t uid = get_parameter(0);

  int status = seteuid(uid);
  write_status32(status);
}


static void
do_pathconf(void)
{
  EMUL_ADDR path = get_parameter(0);
  int name = get_parameter(1);

  int status = pathconf(read_path(path), name);

  write_status32(status);
}


static void
do_fpathconf(void)
{
  EMUL_ADDR fd = get_parameter(0);
  int name = get_parameter(1);

  int status = fpathconf(fd, name);

  write_status32(status);
}


static void
do_getrlimit(void)  /* XXX */
{
  int resource = get_parameter(0);
  EMUL_ADDR rlp = get_parameter(1);

  struct rlimit buf;

  int status = getrlimit(resource, &buf);
  if (status != -1)
    write_memory(rlp, &buf, sizeof(buf));  /* XXX */

  write_status32(status);
}


static void
do_setrlimit(void)  /* XXX */
{
  /* XXX Not all limits are sensible to set within the emulator. */
  write_status32(0);
}


static void
do_mmap(void)  /* XXX */
{
  void* status;
  char* tmp_status;

  void* allocated = NULL;

  /* XXX mmap for new memory. */
  int tmp_len;
  unsigned int tmp_membase;

  EMUL_ADDR addr = get_parameter(0);
  size_t len = get_parameter(1);
  int prot = get_parameter(2);
  int flags = get_parameter(3);
  int fd = get_parameter(4);
  off_t offset = ((off_t)get_parameter(7) << 32) | get_parameter(6);

  if (addr == 0 &&
      prot == (PROT_READ | PROT_WRITE) &&
      flags == (MAP_ANON | MAP_PRIVATE) &&
      fd == -1 &&
      offset == 0)
    {
      int i;

      write_status32(membase);

      /* Round to next page. */
      len = (len + 0xfff) & ~0xfff;

      /* mmap returns zeroed memory. */
      /* XXX This should be done in the page-allocation code instead. */
      for (i = 0; i < len; i++)
        WRITE_INT8(membase + i, 0);

      membase += len;

      return;
    }

  if (fd < 0)
    {
      status = mmap(0, len, prot, flags, fd, offset);
    }
  else
    {
      off_t x;

      x = lseek(fd, offset, SEEK_SET);
      if (x != -1)
        {
          ssize_t y;
          status = xmalloc(len);
          y = read(fd, status, len);
          if (y == -1)
            status = MAP_FAILED;
          else
            allocated = status;
        }
      else
        {
          status = MAP_FAILED;
        }
    }

  if (status == MAP_FAILED)
    {
      /* XXX different ports may have different MAP_FAILED? */
      write_status32(-1);
      return;
    }

  if (len > 4096)
    {
#if 0
      /* We do not handle most succeeded calls yet. */
      fprintf(tfile, "warning mmap %x %x %x %x %x %llx\n",
              addr, len, prot, flags, fd, offset);
#endif
    }

  /* XXX ld.elf_so need us to honor its requested addr. */
  if (addr != 0)
    membase = addr;

  tmp_len = len;
  tmp_membase = membase;
  tmp_status = status;
  while (tmp_len > 0)
    {
      /*      KWARMul_SetPage(emul_state, tmp_membase, tmp_status);*/

      tmp_membase += 4096;
      tmp_len -= 4096;
      tmp_status += 4096;
    }
  write_memory(membase, status, len);
  
  {
    int i;
    size_t len_aligned = (len + 0xfff) & ~0xfff;

    /* mmap returns zeroed memory. */
    /* XXX This should be done in the page-allocation code instead. */
    for (i = len; i < len_aligned; i++)
      WRITE_INT8(membase + i, 0);
  }

  mmap_table[0].addr = status;
  mmap_table[0].eaddr = membase;
  mmap_table[0].neaddr = tmp_membase;
  mmap_table[0].len = len;

  write_status32(membase);

  membase = tmp_membase;

  free(allocated);
}


static void
do_lseek(void)  /* XXX */
{
  int fildes = get_parameter(0);
  off_t offset = ((off_t)get_parameter(3) << 32) | get_parameter(2);
  int whence = get_parameter(4);

  off_t status = lseek(fildes, offset, whence);

  write_status64(status);
}


static void
do_truncate(void)  /* XXX */
{
  EMUL_ADDR path = get_parameter(0);
  off_t length = ((off_t)get_parameter(3) << 32) | get_parameter(2);

  int status = truncate(read_path(path), length);

  write_status32(status);
}


static void
do_ftruncate(void)  /* XXX */
{
  int fd = get_parameter(0);
  off_t length = ((off_t)get_parameter(3) << 32) | get_parameter(2);

  int status = ftruncate(fd, length);

  write_status32(status);
}


static void
do__sysctl(void)  /* XXX */
{
  int i;
  int status = 0; // XXX

  int tmp_name[CTL_MAXNAME];
  unsigned int namelen;
  EMUL_ADDR name;
  EMUL_ADDR oldp;
  EMUL_ADDR oldlenp;
  EMUL_ADDR newp;
  size_t newlen;

  void* tmp_oldp = NULL;
  void* tmp_newp = NULL;
  size_t* tmp_oldlenp = NULL;
  size_t tmp_oldlen;

  name = get_parameter(0);
  namelen = get_parameter(1);
  oldp = get_parameter(2);
  oldlenp = get_parameter(3);
  newp = get_parameter(4);
  newlen = get_parameter(5);

  if (namelen < 2 || namelen > CTL_MAXNAME)
    {
      errno = EINVAL;
      write_status32(0);  // XXX
      return;
    }

  for (i = 0; i < CTL_MAXNAME; i++)
    tmp_name[i] = READ_INT32(name + i * sizeof(EMUL_ADDR));

  if (oldp != 0 && oldlenp == 0)
    {
      errno = EFAULT;
      write_status32(0); // XXX
      return;
    }

  if (oldlenp != 0)
    {
      tmp_oldlen = READ_INT32(oldlenp);
      tmp_oldlenp = &tmp_oldlen;
    }
  if (oldp != 0)
    tmp_oldp = xmalloc(tmp_oldlen);
  if (newp != 0)
    {
      tmp_newp = xmalloc(newlen);
      read_memory(tmp_newp, newp, newlen);
    }

  if (tmp_name[0] == CTL_HW)
    {
      unsigned char buf[256];  /* XXX */
      int intbuf;
      size_t bufsize = 0;

      switch(tmp_name[1])
        {
        case HW_MACHINE:
          strcpy(buf, emul_hw_machine);
          bufsize = strlen(buf) + 1;
          break;
        case HW_MODEL:
          strcpy(buf, EMUL_HW_MODEL);
          bufsize = strlen(buf) + 1;
          break;
        case HW_NCPU:
          intbuf = 1;
          break;
        case HW_BYTEORDER:
          intbuf = EMUL_HW_BYTEORDER;
          break;
        case HW_PHYSMEM:
        case HW_USERMEM:
          {
            size_t size = sizeof(intbuf);
            status = sysctl(tmp_name, namelen,
                            &intbuf, &size,
                            NULL, 0);
          }
          break;
        case HW_PAGESIZE:
          intbuf = EMUL_HW_PAGESIZE;
          break;
        case HW_DISKNAMES:
          xfatal("sysctl HW_DISKNAMES not implemented\n");  // XXX
        case HW_DISKSTATS:
          xfatal("sysctl HW_DISKSTATS not implemented\n");  // XXX
        case HW_MACHINE_ARCH:
          strcpy(buf, EMUL_HW_MACHINE_ARCH);
          bufsize = strlen(buf) + 1;
          break;
        case HW_ALIGNBYTES:
          intbuf = EMUL_HW_ALIGNBYTES;
          break;
        case HW_CNMAGIC:
          xfatal("sysctl HW_CNMAGIC not implemented\n");  // XXX
        default:
          break;
        }

      if (status == -1)
        {
          /* Nothing to do.  errno is already set. */
        }
      else if (newp != 0)
        {
          /* The only changeable item in the HW MIB is HW_CNMAGIC, which
           * we do not support anyway... */
          errno = EPERM;
        }
      else if (bufsize > 0)
        {
          if (oldp != 0)
            {
              write_memory(oldp, buf, MIN(bufsize, tmp_oldlen));
              if (tmp_oldlen < bufsize)
                errno = ENOMEM;
              WRITE_INT32(oldlenp, MIN(bufsize, tmp_oldlen));
            }
          else
            WRITE_INT32(oldlenp, bufsize);
        }
      else
        {
          if (oldp != 0)
            {
              if (tmp_oldlen < 4)
                errno = ENOMEM;
              else
                {
                  WRITE_INT32(oldp, intbuf);
                  WRITE_INT32(oldlenp, 4);
                }
            }
          else
            WRITE_INT32(oldlenp, bufsize);
        }
    }
  else if (tmp_name[0] == CTL_KERN)
    {
      switch(tmp_name[1])
        {
        case KERN_ARGMAX:
        case KERN_AUTONICETIME:
        case KERN_AUTONICEVAL:
        case KERN_CCPU:
        case KERN_FSCALE:
        case KERN_FSYNC:
        case KERN_HOSTID:
        case KERN_IOV_MAX:
        case KERN_JOB_CONTROL:
        case KERN_LOGIN_NAME_MAX:
        case KERN_LOGSIGEXIT:
        case KERN_MAPPED_FILES:
        case KERN_MAXFILES:
        case KERN_MAXPARTITIONS:
        case KERN_MAXPROC:
        case KERN_MAXPTYS:
        case KERN_MAXVNODES:
        case KERN_MEMLOCK:
        case KERN_MEMLOCK_RANGE:
        case KERN_MEMORY_PROTECTION:
        case KERN_MONOTONIC_CLOCK:
        case KERN_MSGBUFSIZE:
        case KERN_NGROUPS:
        case KERN_OSREV:
        case KERN_POSIX1:
        case KERN_RAWPARTITION:
        case KERN_RTC_OFFSET:
        case KERN_SAVED_IDS:
        case KERN_SECURELVL:
        case KERN_SYNCHRONIZED_IO:
        case KERN_SYSVMSG:
        case KERN_SYSVSEM:
        case KERN_SYSVSHM:
          /* int */
          {
            status = sysctl(tmp_name, namelen,
                            tmp_oldp, tmp_oldlenp,
                            tmp_newp, newlen);
            if (status != -1 && oldp != 0)
              {
                int tmp;

                memcpy(&tmp, tmp_oldp, 4);
                WRITE_INT32(oldp, tmp);
                WRITE_INT32(oldlenp, tmp_oldlen);
              }
          }
          break;

        case KERN_OSRELEASE:
          {
            unsigned char buf[256];  /* XXX */
            size_t bufsize = 0;

            strcpy(buf, EMUL_KERN_OSRELEASE);
            bufsize = strlen(buf) + 1;

            if (oldp != 0)
              {
                write_memory(oldp, buf, MIN(bufsize, tmp_oldlen));
                if (tmp_oldlen < bufsize)
                  errno = ENOMEM;
                WRITE_INT32(oldlenp, MIN(bufsize, tmp_oldlen));
              }
            else
              WRITE_INT32(oldlenp, bufsize);
          }
          break;

        case KERN_DEFCORENAME:
        case KERN_DOMAINNAME:
        case KERN_HOSTNAME:
        case KERN_OSTYPE:
        case KERN_PROC_ARGS:
        case KERN_ROOT_DEVICE:
        case KERN_VERSION:
          /* string */
          status = sysctl(tmp_name, namelen,
                          tmp_oldp, tmp_oldlenp,
                          tmp_newp, newlen);
          if (status != -1 && oldp != 0)
            {
              WRITE_INT32(oldlenp, tmp_oldlen);
              write_memory(oldp, tmp_oldp, tmp_oldlen);
            }
          break;

        case KERN_PROC2: // XXX struct kinfoporoc2
          {
            status = sysctl(tmp_name, namelen,
                            tmp_oldp, tmp_oldlenp,
                            tmp_newp, newlen);
            if (status != -1 && oldp != 0)
              {
                write_memory(oldp, tmp_oldp, tmp_oldlen); /* XXX */
                WRITE_INT32(oldlenp, tmp_oldlen);
              }
            else if (oldp == 0 && oldlenp != 0)
              WRITE_INT32(oldlenp, tmp_oldlen);
          }
          break;

        case KERN_BOOTTIME: // XXX struct timeval
          {
            status = sysctl(tmp_name, namelen,
                            tmp_oldp, tmp_oldlenp,
                            tmp_newp, newlen);
            if (status != -1 && oldp != 0)
              {
                WRITE_INT32(oldlenp, tmp_oldlen);
                write_memory(oldp, tmp_oldp, tmp_oldlen);
              }
          }
          break;

        case KERN_CLOCKRATE: // XXX struct clocinfo
        case KERN_CP_TIME: // XXX long[]
        case KERN_FILE: // XXX struct file
        case KERN_MBUF:  // XXX node
        case KERN_MSGBUF: // XXX char[]
        case KERN_NTPTIME: // XXX struct ntptimeval
        case KERN_PROC: // XXX struct kinfo_proc
        case KERN_PROF:  // XXX node
        case KERN_SYSVIPC_INFO: // XXX node
        case KERN_TKSTAT: // XXX node
        case KERN_VNODE: // XXX struct vnode
          xfatal("sysctl CTL_KERN:0x%x not implemented\n", tmp_name[1]);

        default:
          errno = EOPNOTSUPP;
          break;
        }
    }
  else if (tmp_name[0] == CTL_VM)
    {
      switch(tmp_name[1])
        {
        case VM_ANONMIN:
        case VM_ANONMAX:
        case VM_EXECMIN:
        case VM_EXECMAX:
        case VM_FILEMIN:
        case VM_FILEMAX:
        case VM_NKMEMPAGES:
        case VM_MAXSLP:
        case VM_USPACE:
          /* int */
          {
            status = sysctl(tmp_name, namelen,
                            tmp_oldp, tmp_oldlenp,
                            tmp_newp, newlen);
            if (status != -1 && oldp != 0)
              {
                int tmp;

                memcpy(&tmp, tmp_oldp, 4);
                WRITE_INT32(oldp, tmp);
                WRITE_INT32(oldlenp, tmp_oldlen);
              }
          }
          break;

        case VM_LOADAVG:
          {
            status = sysctl(tmp_name, namelen,
                            tmp_oldp, tmp_oldlenp,
                            tmp_newp, newlen);
            if (status != -1 && oldp != 0)
              {
                WRITE_INT32(oldlenp, tmp_oldlen);
                write_memory(oldp, tmp_oldp, tmp_oldlen);
              }
          }
          break;

        case VM_METER:
        case VM_UVMEXP:
        case VM_UVMEXP2:
        default:
          xfatal("sysctl CTL_VM:0x%x not implemented\n", tmp_name[1]);  // XXX
          break;
        }
    }
  else if (tmp_name[0] == CTL_NET)
    {
      xfatal("sysctl CTL_NET:0x%x not implemented\n", tmp_name[1]);  // XXX
    }
  else
    xfatal("sysctl tmp_name[0] = 0x%x not implemented\n", tmp_name[0]);  // XXX

  write_status32(status);  // XXX

  free(tmp_newp);
  free(tmp_oldp);
}


static void
do_clock_gettime(void)
{
  int status;

  clockid_t clock_id = get_parameter(0);
  EMUL_ADDR tp = get_parameter(1);

  struct timespec tmp_tp;

  status = clock_gettime(clock_id, &tmp_tp);
  write_status32(status);
  if (status != -1)
    write_struct_timespec(tp, &tmp_tp);
}


static void
do_clock_settime(void)
{
  int status;

  clockid_t clock_id = get_parameter(0);
  EMUL_ADDR tp = get_parameter(1);

  struct timespec tmp_tp;

  read_struct_timespec(&tmp_tp, tp);

  status = clock_gettime(clock_id, &tmp_tp);
  write_status32(status);
}


static void
do_clock_getres(void)
{
  int status;

  clockid_t clock_id = get_parameter(0);
  EMUL_ADDR res = get_parameter(1);

  struct timespec tmp_res;

  status = clock_gettime(clock_id, &tmp_res);
  write_status32(status);
  if (status != -1)
    write_struct_timespec(res, &tmp_res);
}


static void
do_nanosleep(void)
{
  int status;

  EMUL_ADDR rqtp = get_parameter(0);
  EMUL_ADDR rmtp = get_parameter(1);
  struct timespec tmp_rqtp, tmp_rmtp;

  read_struct_timespec(&tmp_rqtp, rqtp);

  status = nanosleep(&tmp_rqtp, &tmp_rmtp);
  write_status32(status);
  if (status != -1 && rmtp != 0)
    write_struct_timespec(rmtp, &tmp_rmtp);
}


static void
do_mlockall(void)  /* XXX */
{
  /* XXX There is no sensible thing to do in this syscall. */
  write_status32(0);
}


static void
do_munlockall(void)  /* XXX */
{
  /* XXX There is no sensible thing to do in this syscall. */
  write_status32(0);
}


static void
do_mlock(void)  /* XXX */
{
  /* XXX There is no sensible thing to do in this syscall. */
  write_status32(0);
}


static void
do_munlock(void)  /* XXX */
{
  /* XXX There is no sensible thing to do in this syscall. */
  write_status32(0);
}


static void
do_futimes(void)
{
  int status;

  int fd = get_parameter(0);
  EMUL_ADDR times = get_parameter(1);

  struct timeval tmp_buf[2];
  void *tmp_times = NULL;
  if (times != 0)
    {
      times = read_struct_timeval(&tmp_buf[0], times);
      (void)  read_struct_timeval(&tmp_buf[1], times);
      tmp_times = &tmp_buf[0];
    }

  status = futimes(fd, tmp_times);
  write_status32(status);  
}


static void
do_getpgid(void)
{
  pid_t pid = get_parameter(0);

  int status = getpgid(pid);

  write_status32(status);
}


static void
do_poll(void) /* XXX */
{
  int status, i;

  EMUL_ADDR fds = get_parameter(0);
  nfds_t nfds = get_parameter(1);
  int timeout = get_parameter(2);
  EMUL_ADDR fds2 = fds;

  struct pollfd *tmp_fds = xmalloc(nfds * sizeof(*tmp_fds));

  for (i = 0; i < nfds; i++)
    {
      tmp_fds[i].fd = READ_INT32(fds); fds += 4;
      tmp_fds[i].events = READ_INT16(fds); fds += 2;
      tmp_fds[i].revents = READ_INT16(fds); fds += 2;
    }

  status = poll(tmp_fds, nfds, timeout);
  write_status32(status);
  if (status != -1)
    for (i = 0; i < nfds; i++)
      {
        WRITE_INT32(fds2, tmp_fds[i].fd); fds2 += 4;
        WRITE_INT16(fds2, tmp_fds[i].events); fds2 += 2;
        WRITE_INT16(fds2, tmp_fds[i].revents); fds2 += 2;
      }

  free(tmp_fds);
}


static void
do_semget(void)
{
  key_t key = get_parameter(0);
  int nsems = get_parameter(1);
  int semflg = get_parameter(2);

  int status = semget(key, nsems, semflg);

  write_status32(status);
}


static void
do___posix_rename(void)
{
  EMUL_ADDR from = get_parameter(0);
  EMUL_ADDR to = get_parameter(1);

  /* read_path() returns a static bufer, so we cannot call it for both
   * parameters in the function call. */
  char* tmp_from = xstrdup(read_path(from));
  int status = __posix_rename(tmp_from, read_path(to));
  free(tmp_from);

  write_status32(status);
}


static void
do_getdents(void)
{
  int fd = get_parameter(0);
  EMUL_ADDR buf = get_parameter(1);
  size_t nbytes = get_parameter(2);

  void* tmp_buf = xmalloc(nbytes);

  int status = getdents(fd, tmp_buf, nbytes);
  write_status32(status);
  if (status != -1)
    {
      char *cur = tmp_buf;
      char *end = cur + status;
      while (cur < end)
        {
          struct dirent *p = (struct dirent *)cur;

          WRITE_INT32(buf, p->d_fileno);
          WRITE_INT16(buf + 4, p->d_reclen);
          WRITE_INT8(buf + 6, p->d_type);
          WRITE_INT8(buf + 7, p->d_namlen);
          write_memory(buf + 8, p->d_name, p->d_namlen + 1);
          buf += p->d_reclen;
          cur += p->d_reclen;
        }
    }

  free(tmp_buf);
}


static void
do_lchmod(void)
{
  EMUL_ADDR path = get_parameter(0);
  mode_t mode = get_parameter(1);

  int status = lchmod(read_path(path), mode);
  write_status32(status);  
}


static void
do_lchown(void)
{
  EMUL_ADDR path = get_parameter(0);
  uid_t owner = get_parameter(1);
  gid_t group = get_parameter(2);

  int status = lchown(read_path(path), owner, group);

  write_status32(status);
}


static void
do_lutimes(void)
{
  int status;

  EMUL_ADDR path = get_parameter(0);
  EMUL_ADDR times = get_parameter(1);

  struct timeval tmp_buf[2];
  void *tmp_times = NULL;
  if (times != 0)
    {
      times = read_struct_timeval(&tmp_buf[0], times);
      (void)  read_struct_timeval(&tmp_buf[1], times);
      tmp_times = &tmp_buf[0];
    }

  status = lutimes(read_path(path), tmp_times);
  write_status32(status);  
}


static void
do___stat13(void)
{
  EMUL_ADDR path = get_parameter(0);
  EMUL_ADDR sb = get_parameter(1);
  struct stat buf;

  int status = stat(read_path(path), &buf);

  write_status32(status);
  if (status != -1)
    write_struct_stat(sb, &buf);
}


static void
do___fstat13(void)
{
  int fd = get_parameter(0);
  EMUL_ADDR sb = get_parameter(1);
  struct stat buf;

  int status = fstat(fd, &buf);

  write_status32(status);
  if (status != -1)
    write_struct_stat(sb, &buf);
}


static void
do___lstat13(void)
{
  EMUL_ADDR path = get_parameter(0);
  EMUL_ADDR sb = get_parameter(1);
  struct stat buf;

  int status = lstat(read_path(path), &buf);

  write_status32(status);
  if (status != -1)
    write_struct_stat(sb, &buf);
}


static void
do___sigaltstack14(void)  /* XXX */
{
  EMUL_ADDR ss = get_parameter(0);
  EMUL_ADDR oss = get_parameter(1);

  if (ss != 0 &&
      running_signal &&
      emul_state->Reg[13] <= ss_sp &&
      emul_state->Reg[13] > (ss_sp + ss_size))
    {
      /* We are running on the stack we are trying to change. */
      errno = EINVAL;
      write_status32(-1);

      return;
    }

  if (oss != 0)
    {
      int flags;

      flags &= ~SS_ONSTACK;
      if (running_signal &&
          emul_state->Reg[13] <= ss_sp &&
          emul_state->Reg[13] > (ss_sp + ss_size))
        flags |= SS_ONSTACK;

      WRITE_INT32(oss,     ss_sp);
      WRITE_INT32(oss + 4, ss_size);
      WRITE_INT32(oss + 8, flags);
    }

  if (ss != 0)
    {
      ss_sp = READ_INT32(ss);
      ss_size = READ_INT32(ss + 4);
      ss_flags = (READ_INT32(ss + 8)) & ~SS_ONSTACK;
    }

  write_status32(0);
}


static void
do___vfork14(void)
{
  /* XXX This is WRONG!. */
  int status = fork();
  if (status == 0)
    write_status32_2(getppid(), 1);  /* Child */
  else
    write_status32_2(status, 0);  /* Parent */
}


static void
do_getsid(void)
{
  pid_t pid = get_parameter(0);

  int status = getsid(pid);

  write_status32(status);
}


static void
dummy_sig_handler(int sig)
{
  if (pending_sig != -1)
    xfatal("Too many signals!\n");

  pending_sig_handler = signals[sig].handler;
  pending_sig = sig;
}


static void
do___sigaction14(void)  /* XXX */
{
  int status;

  int sig = get_parameter(0);
  EMUL_ADDR act = get_parameter(1);
  EMUL_ADDR oact = get_parameter(2);

  struct sigaction buf_act;
  struct sigaction buf_oact;

  struct sigaction *tmp_act = NULL;
  struct sigaction *tmp_oact = NULL;

  if (act != 0)
    {
      read_memory(&buf_act, act, sizeof(buf_act)); /* XXX */

      if (buf_act.sa_handler != SIG_DFL && buf_act.sa_handler != SIG_IGN)
        {
          signals[sig].handler = READ_INT32(act);
          signals[sig].flags = buf_act.sa_flags;
          buf_act.sa_handler = dummy_sig_handler;
        }
      tmp_act = &buf_act;
    }
  if (oact != 0)
      tmp_oact = &buf_oact;

  status = sigaction(sig, tmp_act, tmp_oact);
  if (status != -1 && tmp_oact != NULL)
    {
      /* XXX Need to translate user's handler adde. */
      write_memory(oact, &buf_oact, sizeof(buf_oact)); /* XXX */
    }

  write_status32(status);
}


static void
do___sigpending14(void)  /* XXX */
{
  EMUL_ADDR set = get_parameter(0);

  sigset_t set_buf;

  int status = sigpending(&set_buf);
  if (status != -1)
    write_memory(set, &set_buf, sizeof(set_buf));

  write_status32(status);
}


static void
do___sigprocmask14(void)  /* XXX */
{
  int status;

  int how = get_parameter(0);
  EMUL_ADDR set = get_parameter(1);
  EMUL_ADDR oset = get_parameter(2);

  sigset_t set_buf, oset_buf;
  sigset_t *tmp_set = NULL;
  sigset_t *tmp_oset = NULL;

  if (set != 0)
    {
      read_memory(&set_buf, set, sizeof(set_buf));
      tmp_set = &set_buf;
    }
  if (oset != 0)
    tmp_oset = &oset_buf;

  status = sigprocmask(how, tmp_set, tmp_oset);
  if (status != -1 && tmp_oset != NULL)
    write_memory(oset, tmp_oset, sizeof(*tmp_oset));

  write_status32(status);
}


static void
do___sigsuspend14(void)  /* XXX */
{
  int status;

  EMUL_ADDR sigmask = get_parameter(0);

  sigset_t tmp_sigmask;

  read_memory(&tmp_sigmask, sigmask, sizeof(tmp_sigmask)); /* XXX */

  status = sigsuspend(&tmp_sigmask);
  write_status32(status);
}


static void
do___sigreturn14(void)  /* XXX */
{
  memcpy(&emul_state->Reg[0], stored_regs, 16*4);
  ARMul_SetCPSR(emul_state, stored_cpsr);

  emul_state->Reg[15] -= 8;

  emul_state->pc = emul_state->Reg[15];
  emul_state->NextInstr = RESUME;

  running_signal = 0;
}


static void
do___getcwd(void)  /* XXX */
{
  EMUL_ADDR buf = get_parameter(0);
  size_t bufsize = get_parameter(1);

  void* tmp_buf = xmalloc(bufsize);
  int status = __getcwd(tmp_buf, bufsize);
  write_status32(status);
  if (status != -1)
    write_path(buf, tmp_buf);

  free(tmp_buf);
}


static void
do_issetugid(void)  /* XXX */
{
  write_status32(0);
}


static void
do_lchflags(void)
{
  EMUL_ADDR path = get_parameter(0);
  u_long flags = get_parameter(1);

  int status = lchflags(read_path(path), flags);

  write_status32(status);
}


static struct {
  void (*handler)(void);
  const char *name;
  int nofargs;
} netbsd_descriptors[] = {
  /*   0 */ {NULL, "syscall", 0},
  /*   1 */ {do_exit, "exit", 1},
  /*   2 */ {do_fork, "fork", 0},
  /*   3 */ {do_read, "read", 3},
  /*   4 */ {do_write, "write", 3},
  /*   5 */ {do_open, "open", 3},
  /*   6 */ {do_close, "close", 1},
  /*   7 */ {do_wait4, "wait4", 4},
  /*   8 */ {NULL, "compat_43_ocreat", 0},
  /*   9 */ {do_link, "link", 2},
  /*  10 */ {do_unlink, "unlink", 1},
  /*  11 */ {NULL, "obsolete execv", 0},
  /*  12 */ {do_chdir, "chdir", 1},
  /*  13 */ {do_fchdir, "fchdir", 1},
  /*  14 */ {do_mknod, "mknod", 3},
  /*  15 */ {do_chmod, "chmod", 2},
  /*  16 */ {do_chown, "chown", 3},
  /*  17 */ {do_break, "break", 1},
  /*  18 */ {do_getfsstat, "getfsstat", 3},
  /*  19 */ {NULL, "compat_43_olseek", 0},
  /*  20 */ {do_getpid, "getpid", 0},
  /*  21 */ {NULL, "mount", 0},
  /*  22 */ {NULL, "unmount", 0},
  /*  23 */ {do_setuid, "setuid", 1},
  /*  24 */ {do_getuid, "getuid", 0},
  /*  25 */ {do_geteuid, "geteuid", 0},
  /*  26 */ {NULL, "ptrace", 0},
  /*  27 */ {NULL, "recvmsg", 0},
  /*  28 */ {NULL, "sendmsg", 0},
  /*  29 */ {do_recvfrom, "recvfrom", 6},
  /*  30 */ {NULL, "accept", 0},
  /*  31 */ {NULL, "getpeername", 0},
  /*  32 */ {NULL, "getsockname", 0},
  /*  33 */ {do_access, "access", 2},
  /*  34 */ {do_chflags, "chflags", 2},
  /*  35 */ {do_fchflags, "fchflags", 2},
  /*  36 */ {do_sync, "sync", 0},
  /*  37 */ {do_kill, "kill", 2},
  /*  38 */ {NULL, "compat_43_stat43", 0},
  /*  39 */ {do_getppid, "getppid", 0},
  /*  40 */ {NULL, "compat_43_lstat43", 0},
  /*  41 */ {do_dup, "dup", 1},
  /*  42 */ {do_pipe, "pipe", 1},
  /*  43 */ {do_getegid, "getegid", 0},
  /*  44 */ {NULL, "profil", 0},
  /*  45 */ {NULL, "ktrace", 0},
  /*  46 */ {NULL, "compat_13_sigaction13", 0},
  /*  47 */ {do_getgid, "getgid", 0},
  /*  48 */ {NULL, "compat_13_sigprocmask13", 0},
  /*  49 */ {do___getlogin, "__getlogin", 2},
  /*  50 */ {NULL, "setlogin", 0},
  /*  51 */ {NULL, "acct", 0},
  /*  52 */ {NULL, "compat_13_sigpending13", 0},
  /*  53 */ {NULL, "compat_13_sigaltstack13", 0},
  /*  54 */ {do_ioctl, "ioctl", 3},
  /*  55 */ {NULL, "compat_12_oreboot", 0},
  /*  56 */ {NULL, "revoke", 0},
  /*  57 */ {do_symlink, "symlink", 2},
  /*  58 */ {do_readlink, "readlink", 3},
  /*  59 */ {do_execve, "execve", 3},
  /*  60 */ {do_umask, "umask", 1},
  /*  61 */ {NULL, "chroot", 0},
  /*  62 */ {NULL, "compat_43_fstat43", 0},
  /*  63 */ {NULL, "compat_43_ogetkerninfo", 0},
  /*  64 */ {NULL, "compat_43_ogetpagesize", 0},
  /*  65 */ {NULL, "compat_12_msync", 0},
  /*  66 */ {do_vfork, "vfork", 0},
  /*  67 */ {NULL, "obsolete vread", 0},
  /*  68 */ {NULL, " is obsolete vwrite", 0},
  /*  69 */ {NULL, "sbrk", 0},
  /*  70 */ {NULL, "sstk", 0},
  /*  71 */ {NULL, "compat_43_ommap", 0},
  /*  72 */ {do_vadvise, "vadvise", 1},
  /*  73 */ {do_munmap, "munmap", 2},
  /*  74 */ {do_mprotect, "mprotect", 3},
  /*  75 */ {do_madvise, "madvise", 3},
  /*  76 */ {NULL, "obsolete vhangup", 0},
  /*  77 */ {NULL, "obsolete vlimit", 0},
  /*  78 */ {NULL, "mincore", 0},
  /*  79 */ {do_getgroups, "getgroups", 2},
  /*  80 */ {do_setgroups, "setgroups", 2},
  /*  81 */ {do_getpgrp, "getpgrp", 0},
  /*  82 */ {do_setpgid, "setpgid", 2},
  /*  83 */ {do_setitimer, "setitimer", 3},
  /*  84 */ {NULL, "compat_43_owait", 0},
  /*  85 */ {NULL, "compat_12_oswapon", 0},
  /*  86 */ {NULL, "getitimer", 0},
  /*  87 */ {NULL, "compat_43_ogethostname", 0},
  /*  88 */ {NULL, "compat_43_osethostname", 0},
  /*  89 */ {NULL, "compat_43_ogetdtablesize", 0},
  /*  90 */ {do_dup2, "dup2", 2},
  /*  91 */ {NULL, NULL, 0},
  /*  92 */ {do_fcntl, "fcntl", 3},
  /*  93 */ {do_select, "select", 5},
  /*  94 */ {NULL, NULL, 0},
  /*  95 */ {do_fsync, "fsync", 1},
  /*  96 */ {do_setpriority, "setpriority", 3},
  /*  97 */ {do_socket, "socket", 3},
  /*  98 */ {do_connect, "connect", 3},
  /*  99 */ {NULL, "compat_43_oaccept", 0},
  /* 100 */ {do_getpriority, "getpriority", 2},
  /* 101 */ {NULL, "compat_43_osend", 0},
  /* 102 */ {NULL, "compat_43_orecv", 0},
  /* 103 */ {NULL, "compat_13_sigreturn13", 0},
  /* 104 */ {do_bind, "bind", 3},
  /* 105 */ {do_setsockopt, "setsockopt", 5},
  /* 106 */ {do_listen, "listen", 2},
  /* 107 */ {NULL, "obsolete vtimes", 0},
  /* 108 */ {NULL, "compat_43_osigvec", 0},
  /* 109 */ {NULL, "compat_43_osigblock", 0},
  /* 110 */ {NULL, "compat_43_osigsetmask", 0},
  /* 111 */ {NULL, "compat_13_sigsuspend13", 0},
  /* 112 */ {NULL, "compat_43_osigstack", 0},
  /* 113 */ {NULL, "compat_43_orecvmsg", 0},
  /* 114 */ {NULL, "compat_43_osendmsg", 0},
  /* 115 */ {NULL, "obsolete vtrace", 0},
  /* 116 */ {do_gettimeofday, "gettimeofday", 2},
  /* 117 */ {do_getrusage, "getrusage", 2},
  /* 118 */ {do_getsockopt, "getsockopt", 5},
  /* 119 */ {NULL, "obsolete resuba", 0},
  /* 120 */ {do_readv, "readv", 3},
  /* 121 */ {do_writev, "writev", 3},
  /* 122 */ {NULL, "settimeofday", 0},
  /* 123 */ {do_fchown, "fchown", 3},
  /* 124 */ {do_fchmod, "fchmod", 2},
  /* 125 */ {NULL, "compat_43_orecvfrom", 0},
  /* 126 */ {do_setreuid, "setreuid", 2},
  /* 127 */ {do_setregid, "setregid", 2},
  /* 128 */ {do_rename, "rename", 2},
  /* 129 */ {NULL, "compat_43_otruncate", 0},
  /* 130 */ {NULL, "compat_43_oftruncate", 0},
  /* 131 */ {do_flock, "flock", 2},
  /* 132 */ {do_mkfifo, "mkfifo", 2},
  /* 133 */ {do_sendto, "sendto", 6},
  /* 134 */ {do_shutdown, "shutdown", 2},
  /* 135 */ {do_socketpair, "socketpair", 4},
  /* 136 */ {do_mkdir, "mkdir", 2},
  /* 137 */ {do_rmdir, "rmdir", 1},
  /* 138 */ {do_utimes, "utimes", 2},
  /* 139 */ {NULL, "obsolete 4.2 sigreturn", 0},
  /* 140 */ {NULL, "adjtime", 0},
  /* 141 */ {NULL, "compat_43_ogetpeername", 0},
  /* 142 */ {NULL, "compat_43_ogethostid", 0},
  /* 143 */ {NULL, "compat_43_osethostid", 0},
  /* 144 */ {NULL, "compat_43_ogetrlimit", 0},
  /* 145 */ {NULL, "compat_43_osetrlimit", 0},
  /* 146 */ {NULL, "compat_43_okillpg", 0},
  /* 147 */ {do_setsid, "setsid", 0},
  /* 148 */ {NULL, "quotactl", 0},
  /* 149 */ {NULL, "compat_43_oquota", 0},
  /* 150 */ {NULL, "compat_43_ogetsockname", 0},
  /* 151 */ {NULL, NULL, 0},
  /* 152 */ {NULL, NULL, 0},
  /* 153 */ {NULL, NULL, 0},
  /* 154 */ {NULL, NULL, 0},
  /* 155 */ {NULL, "nfssvc", 0},
  /* 156 */ {NULL, "compat_43_ogetdirentries", 0},
  /* 157 */ {do_statfs, "statfs", 2},
  /* 158 */ {do_fstatfs, "fstatfs", 2},
  /* 159 */ {NULL, NULL, 0},
  /* 160 */ {NULL, NULL, 0},
  /* 161 */ {NULL, "getfh", 0},
  /* 162 */ {NULL, "compat_09_ogetdomainname", 0},
  /* 163 */ {NULL, "compat_09_osetdomainname", 0},
  /* 164 */ {NULL, "compat_09_ouname", 0},
  /* 165 */ {NULL, "sysarch", 0},
  /* 166 */ {NULL, NULL, 0},
  /* 167 */ {NULL, NULL, 0},
  /* 168 */ {NULL, NULL, 0},
  /* 169 */ {NULL, "compat_10_osemsys", 0},
  /* 170 */ {NULL, "compat_10_omsgsys", 0},
  /* 171 */ {NULL, "compat_10_oshmsys", 0},
  /* 172 */ {NULL, NULL, 0},
  /* 173 */ {do_pread, "pread", 5},
  /* 174 */ {do_pwrite, "pwrite", 5},
  /* 175 */ {NULL, "ntp_gettime", 0},
  /* 176 */ {NULL, "ntp_adjtime", 0},
  /* 177 */ {NULL, NULL, 0},
  /* 178 */ {NULL, NULL, 0},
  /* 179 */ {NULL, NULL, 0},
  /* 180 */ {NULL, NULL, 0},
  /* 181 */ {do_setgid, "setgid", 2},
  /* 182 */ {do_setegid, "setegid", 2},
  /* 183 */ {do_seteuid, "seteuid", 2},
  /* 184 */ {NULL, "lfs_bmapv", 0},
  /* 185 */ {NULL, "lfs_markv", 0},
  /* 186 */ {NULL, "lfs_segclean", 0},
  /* 187 */ {NULL, "lfs_segwait", 0},
  /* 188 */ {NULL, "compat_12_stat12", 0},
  /* 189 */ {NULL, "compat_12_fstat12", 0},
  /* 190 */ {NULL, "compat_12_lstat12", 0},
  /* 191 */ {do_pathconf, "pathconf", 2},
  /* 192 */ {do_fpathconf, "fpathconf", 2},
  /* 193 */ {NULL, NULL, 0},
  /* 194 */ {do_getrlimit, "getrlimit", 2},
  /* 195 */ {do_setrlimit, "setrlimit", 2},
  /* 196 */ {NULL, "compat_12_getdirentries", 0},
  /* 197 */ {do_mmap, "mmap", 8},
  /* 198 */ {NULL, "__syscall", 0},
  /* 199 */ {do_lseek, "lseek", 5},
  /* 200 */ {do_truncate, "truncate", 4},
  /* 201 */ {do_ftruncate, "ftruncate", 4},
  /* 202 */ {do__sysctl, "__sysctl", 6},
  /* 203 */ {do_mlock, "mlock", 2},
  /* 204 */ {do_munlock, "munlock", 2},
  /* 205 */ {NULL, "undelete", 0},
  /* 206 */ {do_futimes, "futimes", 2},
  /* 207 */ {do_getpgid, "getpgid", 1},
  /* 208 */ {NULL, "reboot", 0},
  /* 209 */ {do_poll, "poll", 3},
  /* 210 */ {NULL, NULL, 0},
  /* 211 */ {NULL, NULL, 0},
  /* 212 */ {NULL, NULL, 0},
  /* 213 */ {NULL, NULL, 0},
  /* 214 */ {NULL, NULL, 0},
  /* 215 */ {NULL, NULL, 0},
  /* 216 */ {NULL, NULL, 0},
  /* 217 */ {NULL, NULL, 0},
  /* 218 */ {NULL, NULL, 0},
  /* 219 */ {NULL, NULL, 0},
  /* 220 */ {NULL, "compat_14___semctl"},
  /* 221 */ {do_semget, "semget", 3},
  /* 222 */ {NULL, "semop"},
  /* 223 */ {NULL, "semconfig"},
  /* 224 */ {NULL, "compat_14_msgctl"},
  /* 225 */ {NULL, "msgget"},
  /* 226 */ {NULL, "msgsnd"},
  /* 227 */ {NULL, "msgrcv"},
  /* 228 */ {NULL, "shmat"},
  /* 229 */ {NULL, "compat_14_shmctl"},
  /* 230 */ {NULL, "shmdt"},
  /* 231 */ {NULL, "shmget"},
  /* 232 */ {do_clock_gettime, "clock_gettime", 2},
  /* 233 */ {do_clock_settime, "clock_settime", 2},
  /* 234 */ {do_clock_getres, "clock_getres", 2},
  /* 235 */ {NULL, NULL, 0},
  /* 236 */ {NULL, NULL, 0},
  /* 237 */ {NULL, NULL, 0},
  /* 238 */ {NULL, NULL, 0},
  /* 239 */ {NULL, NULL, 0},
  /* 240 */ {do_nanosleep, "nanosleep", 2},
  /* 241 */ {NULL, "fdatasync", 0},
  /* 242 */ {do_mlockall, "mlockall", 1},
  /* 243 */ {do_munlockall, "munlockall", 0},
  /* 244 */ {NULL, NULL, 0},
  /* 245 */ {NULL, NULL, 0},
  /* 246 */ {NULL, NULL, 0},
  /* 247 */ {NULL, NULL, 0},
  /* 248 */ {NULL, NULL, 0},
  /* 249 */ {NULL, NULL, 0},
  /* 250 */ {NULL, NULL, 0},
  /* 251 */ {NULL, NULL, 0},
  /* 252 */ {NULL, NULL, 0},
  /* 253 */ {NULL, NULL, 0},
  /* 254 */ {NULL, NULL, 0},
  /* 255 */ {NULL, NULL, 0},
  /* 256 */ {NULL, NULL, 0},
  /* 257 */ {NULL, NULL, 0},
  /* 258 */ {NULL, NULL, 0},
  /* 259 */ {NULL, NULL, 0},
  /* 260 */ {NULL, NULL, 0},
  /* 261 */ {NULL, NULL, 0},
  /* 262 */ {NULL, NULL, 0},
  /* 263 */ {NULL, NULL, 0},
  /* 264 */ {NULL, NULL, 0},
  /* 265 */ {NULL, NULL, 0},
  /* 266 */ {NULL, NULL, 0},
  /* 267 */ {NULL, NULL, 0},
  /* 268 */ {NULL, NULL, 0},
  /* 269 */ {NULL, NULL, 0},
  /* 270 */ {do___posix_rename, "__posix_rename", 2},
  /* 271 */ {NULL, "swapctl", 0},
  /* 272 */ {do_getdents, "getdents", 3},
  /* 273 */ {NULL, "minherit", 0},
  /* 274 */ {do_lchmod, "lchmod", 2},
  /* 275 */ {do_lchown, "lchown", 3},
  /* 276 */ {do_lutimes, "lutimes", 2},
  /* 277 */ {NULL, "__msync13", 0},
  /* 278 */ {do___stat13, "__stat13", 2},
  /* 279 */ {do___fstat13, "__fstat13", 2},
  /* 280 */ {do___lstat13, "__lstat13", 2},
  /* 281 */ {do___sigaltstack14, "__sigaltstack14", 2},
  /* 282 */ {do___vfork14, "__vfork14", 0}, /* XXX ? */
  /* 283 */ {NULL, "__posix_chown", 0},
  /* 284 */ {NULL, "__posix_fchown", 0},
  /* 285 */ {NULL, "__posix_lchown", 0},
  /* 286 */ {do_getsid, "getsid", 1},
  /* 287 */ {NULL, "__clone", 0},
  /* 288 */ {NULL, "fktrace", 0},
  /* 289 */ {NULL, "preadv", 0},
  /* 290 */ {NULL, "pwritev", 0},
  /* 291 */ {do___sigaction14, "__sigaction14", 3},
  /* 292 */ {do___sigpending14, "__sigpending14", 1},
  /* 293 */ {do___sigprocmask14, "__sigprocmask14", 3},
  /* 294 */ {do___sigsuspend14, "__sigsuspend14", 1},
  /* 295 */ {do___sigreturn14, "__sigreturn14", 0},
  /* 296 */ {do___getcwd, "__getcwd", 2},
  /* 297 */ {NULL, "fchroot", 0},
  /* 298 */ {NULL, "fhopen", 0},
  /* 299 */ {NULL, "fhstat", 0},
  /* 300 */ {NULL, "fhstatfs", 0},
  /* 301 */ {NULL, "____semctl13", 0},
  /* 302 */ {NULL, "__msgctl13", 0},
  /* 303 */ {NULL, "__shmctl13", 0},
  /* 304 */ {do_lchflags, "lchflags", 2},
  /* 305 */ {do_issetugid, "issetugid", 0},
  /* 306 */ {NULL, "utrace", 0},
};


int
netbsd_syscall(ARMul_State * state, int number)
{
  parameter_offset = 0;

  if ((number & 0xa00000) != 0xa00000)
    xfatal("Unknown SWI number = %d (0x%x)\n", number, number);

  number &= ~0xa00000;

  if (number == 198)  /* __syscall */
    {
      number = get_parameter(0);
      parameter_offset = 2;
    }

  if (number >= sizeof(netbsd_descriptors)/sizeof(*netbsd_descriptors))
    xfatal("Unknown syscall number = %d (0x%x)\n", number, number);

  if (netbsd_descriptors[number].handler == NULL)
    xfatal("Syscall number = %d (%s) is not implemented\n",
           number, netbsd_descriptors[number].name);

  trace_call(netbsd_descriptors[number].name,
             netbsd_descriptors[number].nofargs);

  errno = 0;
  netbsd_descriptors[number].handler();

  return TRUE;
}


void
netbsd_stack_init(ARMul_State * state, int argc, char ** argv)
{
  int i;
  EMUL_ADDR addr, p;
  size_t gap;

  xasprintf(&filesystem_path, "%s/netbsd1-%s", EMUL_BASEDIR, emul_hw_machine);

  emul_state = state;

  /* Point at a space above the stack where to store the strings
   * that arv and envp points to. */
  /* XXX */
  gap = 0x1000 + argc*4;
  addr = ADDRSUPERSTACK + gap;

  p = ADDRSUPERSTACK;

  /* Set up argc/argv. */
  ARMul_WriteWord(state, p, argc);
  p+= 4;
  for (i = 0; i < argc; i++)
    {
      size_t len = strlen(argv[i]) + 1;

      write_memory(addr, argv[i], len);
      ARMul_WriteWord(state, p, addr);

      p += 4;
      addr += len;
    }
  ARMul_WriteWord(state, p, 0); p+= 4;

  /* Set up envp. */
  for (i = 0; environ != NULL && environ[i] != NULL; i++)
    {
      size_t len = strlen(environ[i]) + 1;

      write_memory(addr, environ[i], len);
      ARMul_WriteWord(state, p, addr);

      p += 4;
      addr += len;
    }
  ARMul_WriteWord(state, p, 0); p+= 4;

#if 0
  /* XXX should be at ADDRSUPERSTACK - sizeof(ps_srings)
   * see kern_exec.c and sys/exec.h */
  /* ps_string. */
  ARMul_WriteWord(state, p, 0); p+= 4;
#endif
  {
    EMUL_ADDR phdr;
    Elf32_Ehdr ehdr;

    /* XXX This info can probably be returned from BFD.  Will look
     * into that later... */
    {
      int fd;
      ssize_t len;
      size_t phdr_len;
      Elf32_Phdr* ptr;

      fd = open(netbsd_real_prog, O_RDONLY, 0);
      if (fd == -1)
        xfatal("Cannot open file\n");

      len = read(fd, &ehdr, sizeof(ehdr));
      if (len != sizeof(ehdr))
        xfatal("Read error\n");

      phdr_len = ehdr.e_phnum * sizeof(Elf32_Phdr);
      ptr = xmalloc(phdr_len);

      len = read(fd, ptr, phdr_len);
      if (len != phdr_len)
        xfatal("Read error\n");

      for (i = 0; i < ehdr.e_phnum; i++)
        if (ptr[i].p_type == PT_PHDR)
          {
            /* XXX Need to map it into emu */
            phdr = ptr[i].p_vaddr;
            write_memory(phdr, ptr, phdr_len);
          }

      free(ptr);
    }

    ARMul_WriteWord(state, p, AT_PHDR); p+= 4;
    ARMul_WriteWord(state, p, phdr); p+= 4;

    ARMul_WriteWord(state, p, AT_PHENT); p+= 4;
    ARMul_WriteWord(state, p, ehdr.e_phentsize); p+= 4;

    ARMul_WriteWord(state, p, AT_PHNUM); p+= 4;
    ARMul_WriteWord(state, p, ehdr.e_phnum); p+= 4;

    ARMul_WriteWord(state, p, AT_PAGESZ); p+= 4;
    ARMul_WriteWord(state, p, 4096); p+= 4;

    ARMul_WriteWord(state, p, AT_BASE); p+= 4;
    ARMul_WriteWord(state, p, 0x20018000); p+= 4;

    ARMul_WriteWord(state, p, AT_FLAGS); p+= 4;
    ARMul_WriteWord(state, p, 0); p+= 4;

    ARMul_WriteWord(state, p, AT_ENTRY); p+= 4;
    ARMul_WriteWord(state, p, ehdr.e_entry); p+= 4;

    if (p > (ADDRSUPERSTACK + gap))
      xfatal("Initial stack smashed!\n");
  }

  {
    unsigned char foo[] =
    {
      0x00, 0xe0, 0x8f, 0xe2, /* add     lr, pc, #0 */
      0x03, 0xf0, 0xa0, 0xe1, /* mov     pc, r3 */
      0x10, 0x00, 0x8d, 0xe2, /* add     r0, sp, #16 */
      0x27, 0x01, 0xa0, 0xef, /* swi     0x00000127 */
      0x01, 0x00, 0x00, 0xef, /* swi     0x00000001 */
      0x01, 0x00, 0x00, 0xea  /* b       c <sigcode+0xc> */
    };
    
    /* The signal trampoline. */
    sigcode = (addr + 7) & ~7;
    write_memory(sigcode, foo, 6*4);
  }
}


int
check_signal(void)
{
  if (pending_sig == -1 || running_signal)
    return 0;

  memcpy(stored_regs, &emul_state->Reg[0], 16*4);
  stored_cpsr = ARMul_GetCPSR(emul_state);

  emul_state->Reg[0] = pending_sig;
  emul_state->Reg[1] = 0;  /* XXX Code */
  emul_state->Reg[2] = 0;  /* XXX Scp */
  emul_state->Reg[3] = signals[pending_sig].handler;

  if ((signals[pending_sig].flags & SA_ONSTACK) != 0 &&
      (ss_flags & SS_DISABLE) == 0)
    {
      if (emul_state->Reg[13] > ss_sp ||
          emul_state->Reg[13] <= (ss_sp + ss_size))
        {
          /* We are not running on the signal stack.  Switch to it. */
          emul_state->Reg[13] = ss_sp;
        }
    }

  emul_state->Reg[15] = emul_state->pc = sigcode;
  emul_state->NextInstr = RESUME;

  running_signal = 1;

  pending_sig = -1;

  return 1;
}
